#define _GNU_SOURCE

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>

#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/signal.h>
#include <sys/wait.h>

#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <err.h>
#include <errno.h>

#include "loader.h"
#include "slimfs.h"

typedef struct {
	char root_dir[PATH_MAX];
	struct fuse *fuse;
	bool freed;
} *fuse_pthread_data_t;

static slimfs_t fs;
static char *mount_dir;

static void *fuse_pthread_entry(void *pdata) {
	slimfs_t fs = pdata;

	if (fuse_loop(fs->fuse) < 0) err(1, "Couldn't start fuse loop");

	return NULL;
}

static slimfs_t slimfs_mount(const char *target_dir, const char *root_dir, const char *img_path, size_t img_offset) {
	slimfs_t fs = slimfs_new(1, (char*[]){ "", NULL }, target_dir, root_dir, img_path, img_offset);
	if (fs == NULL) return NULL;

	pthread_t thread;
	pthread_create(&thread, NULL, fuse_pthread_entry, fs);

	return fs;
}

static void handle_close() {
	fprintf(stderr, "\nCTRL+C detected, closing SlimFS...\n");
	slimfs_free(fs);
	rmdir(mount_dir);
	exit(0);
}

static int write_flag(const char *file, const char *format, ...) {
	int fd = open(file, O_WRONLY);
	if (fd < 0) {
		warn("Failed to write to %s", file);
		return -1;
	}

	va_list list;

	va_start(list, format);
	char buff[vsnprintf(NULL, 0, format, list) + 1];
	va_end(list);

	va_start(list, format);
	vsnprintf(buff, sizeof buff, format, list);
	va_end(list);

	int n = write(fd, buff, sizeof buff - 1);
	if (n < 0 || (size_t)n != sizeof buff - 1) {
		warn("Failed to write to %s", file);
		return -1;
	}

	close(fd);
	return 0;
}

static int err_mount(const char *name) {
	warn("Failed to setup virtual mount for %s", name);
	return 127;
}

int slimpak_isolate(const char *root_dir, char **argv) {
	pid_t pid = fork();
	if (pid < 0) err(EXIT_EXECERROR, "fork error");
	else if (pid == 0) {
		int uid = geteuid();
		int gid = getegid();

		errno = 0;

		// Child is responsible for running the actual command
		if (unshare(CLONE_NEWUSER | CLONE_NEWNS) < 0) {
			warn("Failed to containerize");
			exit(127);
		}

		if (write_flag("/proc/self/uid_map", "%d %d 1\n", uid, uid) < 0) exit(127);
		if (write_flag("/proc/self/setgroups", "deny\n") < 0) exit(127);
		if (write_flag("/proc/self/gid_map", "%d %d 1\n", gid, gid) < 0) exit(127);

		char *cwd = getcwd(NULL, 0);
		chdir(root_dir);

		// if (mount(NULL, "/", NULL, MS_REC | MS_SILENT | MS_SLAVE, NULL) < 0) return err_mount("root");
		if (mount("/dev", "dev", NULL, MS_REC | MS_SILENT | MS_BIND, NULL) < 0) return err_mount("/dev");
		if (mount("/proc", "proc", NULL, MS_REC | MS_SILENT | MS_BIND, NULL) < 0) return err_mount("/proc");
		if (mount("/run", "run", NULL, MS_REC | MS_SILENT | MS_BIND, NULL) < 0) return err_mount("/run");

		if (chroot(root_dir) < 0) {
			warn("Failed to chroot into virtual root");
			exit(127);
		}

		chdir(cwd);
		free(cwd);

		execv(argv[0], argv);

		err(127, "Failed to execute command");
	}
	else {
		int status;
		if (waitpid(pid, &status, 0) < 0) {
			warn("Failed to get child status");
			return -1;
		}

		if (WIFEXITED(status)) return (uint16_t)WEXITSTATUS(status);
		else return -1;
	}

	return 0;
}

int main(int argc, char **argv) {
	const char *self_path = "/proc/self/exe";
	const char *temp_base = "/tmp";
	const char *slimpack_arg = NULL;

	size_t offset = slimpack_get_offset(self_path);

	if (argc == 2 && argv[1][0] == '-' && argv[1][1] == '-') {
		slimpack_arg = (const char*)argv[1] + 2;
	}

	if (slimpack_arg && !strcmp(slimpack_arg, "slimpack-offset")) {
		printf("%zu\n", offset);
		return 0;
	}

	mount_dir = malloc(strlen(temp_base) + 19 + 1);
	strcpy(mount_dir, temp_base);
	strcat(mount_dir, "/.slimpack.XXXXXXXX");

	if (!mkdtemp(mount_dir)) err(EXIT_EXECERROR, "create mount dir error");

	fs = slimfs_mount(mount_dir, "/", self_path, offset);
	if (fs == NULL) errx(1, "Couldn't mount SlimFS");

	signal(SIGINT, handle_close);

	int status;

	if (slimpack_arg && !strcmp(slimpack_arg, "slimpack-mount")) {
		fprintf(stderr, "The mountpoint is at %s. Press Ctrl+D when you're ready\n", mount_dir);

		char buff[4096];

		while (read(STDIN_FILENO, buff, sizeof buff) != 0);

		slimfs_free(fs);
		rmdir(mount_dir);

		return 0;
	}

	if (slimpack_arg && !strcmp(slimpack_arg, "slimpack-shell")) {
		status = slimpak_isolate((const char*)mount_dir, (char*[]) { "/bin/bash", NULL });
	}
	else {
		argv[0] = "/entry";
		status = slimpak_isolate((const char*)mount_dir, argv);
	}

	slimfs_free(fs);
	rmdir(mount_dir);

	if (status < 0) {
		err(127, "Failed to start isolated process (status %d)", status);
	}
	else {
		return status;
	}
}
