#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>

#include "loader.h"

#define EI_NIDENT 16

#define ELFCLASS32 1
#define ELFDATA2LSB 1
#define ELFDATA2MSB 2
#define ELFCLASS64 2
#define EI_CLASS 4
#define EI_DATA 5

#if __BYTE_ORDER == __LITTLE_ENDIAN
	#define ELFDATANATIVE ELFDATA2LSB
#elif __BYTE_ORDER == __BIG_ENDIAN
	#define ELFDATANATIVE ELFDATA2MSB
#else
	#error "Unknown machine endian"
#endif

typedef struct {
	unsigned char e_ident[EI_NIDENT];
	uint16_t e_type;
	uint16_t e_machine;
	uint32_t e_version;
	uint32_t e_entry; /* Entry point */
	uint32_t e_phoff;
	uint32_t e_shoff;
	uint32_t e_flags;
	uint16_t e_ehsize;
	uint16_t e_phentsize;
	uint16_t e_phnum;
	uint16_t e_shentsize;
	uint16_t e_shnum;
	uint16_t e_shstrndx;
} elf32_hdr_t;

typedef struct {
	uint32_t sh_name;
	uint32_t sh_type;
	uint32_t sh_flags;
	uint32_t sh_addr;
	uint32_t sh_offset;
	uint32_t sh_size;
	uint32_t sh_link;
	uint32_t sh_info;
	uint32_t sh_addralign;
	uint32_t sh_entsize;
} elf32_shdr_t;

typedef struct {
	unsigned char e_ident[EI_NIDENT]; /* ELF "magic number" */
	uint16_t e_type;
	uint16_t e_machine;
	uint32_t e_version;
	uint64_t e_entry; /* Entry point virtual address */
	uint64_t e_phoff; /* Program header table file offset */
	uint64_t e_shoff; /* Section header table file offset */
	uint32_t e_flags;
	uint16_t e_ehsize;
	uint16_t e_phentsize;
	uint16_t e_phnum;
	uint16_t e_shentsize;
	uint16_t e_shnum;
	uint16_t e_shstrndx;
} elf64_hdr_t;

typedef struct {
	uint32_t sh_name; /* Section name, index in string tbl */
	uint32_t sh_type; /* Type of section */
	uint64_t sh_flags; /* Miscellaneous section attributes */
	uint64_t sh_addr; /* Section virtual addr at execution */
	uint64_t sh_offset; /* Section file offset */
	uint64_t sh_size; /* Size of section in bytes */
	uint32_t sh_link; /* Index of another section */
	uint32_t sh_info; /* Additional section information */
	uint64_t sh_addralign; /* Section alignment */
	uint64_t sh_entsize; /* Entry size if section holds table */
} elf64_shdr_t;

/* Note header in a PT_NOTE section */
typedef struct elf32_note {
	uint32_t n_namesz; /* Name size */
	uint32_t n_descsz; /* Content size */
	uint32_t n_type; /* Content type */
} Elf32_Nhdr;

typedef Elf32_Nhdr Elf_Nhdr;

static uint16_t file16_to_cpu(uint8_t ident[EI_NIDENT], uint16_t val) {
	if (ident[EI_DATA] != ELFDATANATIVE) {
		uint16_t a = val & 0xFF;
		uint16_t b = (val >> 8) & 0xFF;
		val = a << 8 | b;
	}

	return val;
}
static uint32_t file32_to_cpu(uint8_t ident[EI_NIDENT], uint32_t val) {
	if (ident[EI_DATA] != ELFDATANATIVE) {
		uint32_t a = val & 0xFF;
		uint32_t b = (val >> 8) & 0xFF;
		uint32_t c = (val >> 16) & 0xFF;
		uint32_t d = (val >> 24) & 0xFF;
		val = a << 24 | b << 16 | c << 8 | d;
	}
	return val;
}
static uint64_t file64_to_cpu(uint8_t ident[EI_NIDENT], uint64_t val) {
	if (ident[EI_DATA] != ELFDATANATIVE) {
		uint64_t a = val & 0xFF;
		uint64_t b = (val >> 8) & 0xFF;
		uint64_t c = (val >> 16) & 0xFF;
		uint64_t d = (val >> 24) & 0xFF;
		uint64_t e = (val >> 32) & 0xFF;
		uint64_t f = (val >> 40) & 0xFF;
		uint64_t g = (val >> 48) & 0xFF;
		uint64_t h = (val >> 56) & 0xFF;
		val = a << 56 | b << 48 | c << 40 | d << 32 | e << 24 | f << 16 | g << 8 | h;
	}
	return val;
}

static ssize_t read_elf32_size(uint8_t ident[EI_NIDENT], FILE *fd) {
	ssize_t ret;

	fseek(fd, 0, SEEK_SET);

	elf32_hdr_t header;
	ret = fread(&header, 1, sizeof(header), fd);
	if (ret < 0 || ret != sizeof header) {
		warn("Read of ELF header failed");
		return -1;
	}

	size_t sh_offset = file32_to_cpu(ident, header.e_shoff);
	size_t sh_entry_size = file16_to_cpu(ident, header.e_shentsize);
	size_t sh_entry_n = file16_to_cpu(ident, header.e_shnum);

	off_t last_shdr_offset = sh_offset + (sh_entry_size * (sh_entry_n - 1));
	fseek(fd, last_shdr_offset, SEEK_SET);

	elf32_shdr_t sect_header;
	ret = fread(&sect_header, 1, sizeof(sect_header), fd);
	if (ret < 0 || (size_t) ret != sizeof(sect_header)) {
		warn("Read of ELF section header failed");
		return -1;
	}

	/* ELF ends either with the table of section headers (SHT) or with a section. */
	off_t sh_end = sh_offset + (sh_entry_size * sh_entry_n);
	off_t s_end = file64_to_cpu(ident, sect_header.sh_offset) + file64_to_cpu(ident, sect_header.sh_size);

	return sh_end > s_end ? sh_end : s_end;
}
static ssize_t read_elf64_size(uint8_t ident[EI_NIDENT], FILE *fd) {
	off_t ret;

	fseek(fd, 0, SEEK_SET);

	elf64_hdr_t header;
	ret = fread(&header, 1, sizeof(header), fd);
	if (ret < 0 || (size_t) ret != sizeof(header)) {
		warn("Read of ELF header failed");
		return -1;
	}

	size_t sh_offset = file64_to_cpu(ident, header.e_shoff);
	size_t sh_entry_size = file16_to_cpu(ident, header.e_shentsize);
	size_t sh_entry_n = file16_to_cpu(ident, header.e_shnum);

	off_t last_shdr_offset = sh_offset + (sh_entry_size * (sh_entry_n - 1));
	fseek(fd, last_shdr_offset, SEEK_SET);

	elf64_shdr_t sect_header;
	ret = fread(&sect_header, 1, sizeof(sect_header), fd);
	if (ret < 0 || ret != sizeof(sect_header)) {
		warn("Read of ELF section header failed");
		return -1;
	}

	/* ELF ends either with the table of section headers (SHT) or with a section. */
	off_t sht_end = sh_offset + (sh_entry_size * sh_entry_n);
	off_t s_end = file64_to_cpu(ident, sect_header.sh_offset) + file64_to_cpu(ident, sect_header.sh_size);
	return sht_end > s_end ? sht_end : s_end;
}

ssize_t slimpack_get_offset(const char *executable) {
	FILE *fd = fopen(executable, "rb");
	if (fd == NULL) {
		warn("Cannot open executable");
		return -1;
	}

	uint8_t ident[EI_NIDENT];

	off_t ret = fread(ident, 1, EI_NIDENT, fd);
	if (ret != EI_NIDENT) {
		warn("Read of e_ident failed");
		return -1;
	}

	if (ident[EI_DATA] != ELFDATA2LSB && ident[EI_DATA] != ELFDATA2MSB) {
		warnx("Unknown ELF data order %u", ident[EI_DATA]);
		return -1;
	}

	size_t res;

	if (ident[EI_CLASS] == ELFCLASS32) res = read_elf32_size(ident, fd);
	else if (ident[EI_CLASS] == ELFCLASS64) res = read_elf64_size(ident, fd);
	else {
		fprintf(stderr, "Unknown ELF class %u\n", ident[EI_CLASS]);
		return -1;
	}

	fclose(fd);
	return res;
}
