// Structure inspired by Radek Podgorny's unionfs
// License: BSD-style license
// Copyright: Radek Podgorny <radek@podgorny.cz>,
//            Bernd Schubert <bernd-schubert@gmx.de>

#define _GNU_SOURCE

#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <inttypes.h>
#include <libgen.h>
#include <errno.h>
#include <err.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/xattr.h>
#include <sys/vfs.h>
#include <sys/sysmacros.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <fuse.h>
#include <squashfuse/squashfuse.h>
#include <squashfuse/stat.h>

#include "slimfs.h"

#define DEBUG 0

typedef struct {
	struct sqfs_inode sqfs_fd;
	int real_fd;
	bool is_sqfs;
} user_fd_t;

static struct sqfs_inode *find_squashfs_node(struct sqfs *fs, const char *path, struct sqfs_inode *inode) {
	bool found;
	sqfs_err err;

	struct sqfs_inode node;
	if ((err = sqfs_inode_get(fs, &node, fs->sb.root_inode)) != SQFS_OK) {
		warnx("SQFS error during lookup of %s: %d", path, err);
		return NULL;
	}

	if ((err = sqfs_lookup_path(fs, &node, path, &found)) != SQFS_OK) {
		warnx("SQFS error during lookup of %s: %d", path, err);
		return NULL;
	}
	if (!found) return NULL;

	*inode = node;
	return inode;
}

static char *cat_path(char *res, size_t n, const char *root, const char *path) {
	// TODO: check length
	snprintf(res, n, "%s%s", root, path);
	return res;
}

static void *slimfs_init(
	#if FUSE_USE_VERSION < 30
		struct fuse_conn_info *conn
	#else
		struct fuse_conn_info *conn, struct fuse_config *cfg
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS INIT\n");

	#ifdef FUSE_CAP_IOCTL_DIR
		conn->want |= conn->capable & FUSE_CAP_IOCTL_DIR;
	#endif

	slimfs_t data = fuse_get_context()->private_data;
	return data;
}
static void slimfs_destroy(void *pdata) {
	if (DEBUG) fprintf(stderr, "SLIMFS DESTROY\n");
	slimfs_t fs = pdata;

	fuse_unmount(fs->target_dir, fs->channel);
	sqfs_destroy(&fs->image);
	free(fs);
}

static int slimfs_create(const char *path, mode_t mode, struct fuse_file_info *fi) {
	if (DEBUG) fprintf(stderr, "SLIMFS CREATE %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	if (find_squashfs_node(&data->image, path, &inode)) return -EEXIST;
	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	int res = open(buff, fi->flags, mode);
	if (res == -1) return -errno;

	user_fd_t *res_fd = malloc(sizeof *res_fd);
	res_fd->is_sqfs = false;
	res_fd->real_fd = res;

	fi->fh = (uint64_t)res_fd;

	return 0;
}
static int slimfs_symlink(const char *from, const char *to) {
	if (DEBUG) fprintf(stderr, "SLIMFS SYMLINK %s -> %s\n", from, to);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char real_from[PATH_MAX];
	char real_to[PATH_MAX];

	// Same as link, we can't symlink from or to the overlay
	if (find_squashfs_node(&data->image, from, &inode)) return -EROFS;
	if (find_squashfs_node(&data->image, to, &inode)) return -EEXIST;

	if (!cat_path(real_from, sizeof real_from, data->root_dir, from)) return -ENAMETOOLONG;
	if (!cat_path(real_to, sizeof real_to, data->root_dir, to)) return -ENAMETOOLONG;

	if (symlink(real_from, real_to) < 0) return -errno;
	return 0;
}
static int slimfs_link(const char *from, const char *to) {
	if (DEBUG) fprintf(stderr, "SLIMFS LINK %s -> %s\n", from, to);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char real_from[PATH_MAX];
	char real_to[PATH_MAX];

	// We can't hard-link stuff from or to the overlay
	if (find_squashfs_node(&data->image, from, &inode)) return -EROFS;
	if (find_squashfs_node(&data->image, to, &inode)) return -EEXIST;

	if (!cat_path(real_from, sizeof real_from, data->root_dir, from)) return -ENAMETOOLONG;
	if (!cat_path(real_to, sizeof real_to, data->root_dir, to)) return -ENAMETOOLONG;

	if (link(real_from, real_to) < 0) return -errno;
	return 0;
}
static int slimfs_mkdir(const char *path, mode_t mode) {
	if (DEBUG) fprintf(stderr, "SLIMFS MKDIR %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// Already exists in overlay
	if (find_squashfs_node(&data->image, path, &inode)) return -EEXIST;

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	// Let the OS figure out there's a collision
	if (mkdir(buff, mode) < 0) return -errno;
	return 0;
}
static int slimfs_mknod(const char *path, mode_t mode, dev_t rdev) {
	if (DEBUG) fprintf(stderr, "SLIMFS MKNOD %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// Something is already in the overlay
	if (find_squashfs_node(&data->image, path, &inode)) return -EEXIST;

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;
	if (mknod(buff, mode, rdev) < 0) return -errno;
	return 0;

	// int i = find_rw_branch_cutlast(path);
	// if (i == -1) return -errno;

	// char p[PATHLEN_MAX];
	// if (BUILD_PATH(p, uopt.branches[i].path, path)) return -ENAMETOOLONG;

	// int file_type = mode & S_IFMT;
	// int file_perm = mode & (S_PROT_MASK);

	// int res = -1;
	// if (file_type == S_IFREG) {
	// 	// under FreeBSD, only the super-user can create ordinary files using mknod
	// 	// Actually this workaround should not be required any more
	// 	// since we now have the slimfs_create() method
	// 	// So can we remove it?

	// 	usyslog(LOG_INFO, "deprecated mknod workaround, tell the slimfs-fuse authors if you see this!\n");

	// 	res = creat(p, 0);
	// 	if (res > 0 && close(res) == -1) usyslog(LOG_WARNING, "Warning, cannot close file\n");
	// }
	// else {
	// 	res = mknod(p, file_type, rdev);
	// }

	// if (res == -1) return -errno;

	// set_owner(p); // no error check, since creating the file succeeded
	// // NOW, that the file has the proper owner we may set the requested mode
	// chmod(p, file_perm);

	// remove_hidden(path, i);

	// return 0;
}

static int slimfs_rename(
	#if FUSE_USE_VERSION < 30
		const char *from, const char *to
	#else
		const char *from, const char *to, unsigned int flags
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS RENAME %s -> %s\n", from, to);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char real_from[PATH_MAX];
	char real_to[PATH_MAX];

	// We can't move files from or to the overlay, since it's readonly
	if (find_squashfs_node(&data->image, from, &inode)) return -EROFS;
	if (find_squashfs_node(&data->image, to, &inode)) return -EEXIST;

	if (!cat_path(real_from, sizeof real_from, data->root_dir, from)) return -ENAMETOOLONG;
	if (!cat_path(real_to, sizeof real_to, data->root_dir, to)) return -ENAMETOOLONG;

	if (rename(real_from, real_to) < 0) return -errno;
	return 0;
}

static int slimfs_rmdir(const char *path) {
	if (DEBUG) fprintf(stderr, "SLIMFS RMDIR %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// We have two ways we could handle this:
	// 1. If it is in the root, delete it, otherwise, if it's in the overlay, throw a EROFS
	//    The pitfall of this is that if we have the same dir in the root and the overlay,
	//    we will be able to seemingly delete it, but the wrapper will still remain
	// 2. Throw if it exists in the overlay
	//    This is what is done here, the pitfall is that we can't delete a RW directory in
	//    the root, if it's shadowed by the overlay, but is probably what we mean by "rmdir"
	if (find_squashfs_node(&data->image, path, &inode)) return -EROFS;

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	int res = rmdir(buff);
	if (res < 0) return -errno;
	return 0;
}
static int slimfs_unlink(const char *path) {
	if (DEBUG) fprintf(stderr, "SLIMFS RMDIR %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	if (find_squashfs_node(&data->image, path, &inode)) return -EROFS;
	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	int res = unlink(buff);
	if (res < 0) return -errno;
	return 0;
}

static int slimfs_readlink(const char *path, char *buf, size_t size) {
	if (DEBUG) fprintf(stderr, "SLIMFS READLINK %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// If we have a file in the overlay, we can open it only in read-only mode
	if (find_squashfs_node(&data->image, path, &inode)) {
		size_t size = sizeof buff;
		if (sqfs_readlink(&data->image, &inode, buf, &size) != F_OK) {
			return -EINVAL;
		}

		return 0;
	}

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	int res = readlink(buff, buf, size);
	if (res < 0) return -errno;

	buf[res] = '\0';
	return 0;
}
// TODO: BROKEN!!!!
static int slimfs_readdir(
	#if FUSE_USE_VERSION < 30
		const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi
	#else
		const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi, enum fuse_readdir_flags flags
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS READDIR %s\n", path);
	int err = 0;

	slimfs_t data = fuse_get_context()->private_data;
	char buff[PATH_MAX];
	struct sqfs_inode inode;

	char **passed = NULL;
	size_t passed_n = 0;
	size_t passed_i = 0;

	// If we have a hit in the overlay, we will try to iterate it
	if (find_squashfs_node(&data->image, path, &inode)) {
		sqfs_dir dir;
		sqfs_dir_entry entry;
		sqfs_err err;

		if (sqfs_dir_open(&data->image, &inode, &dir, 0) != SQFS_OK) return -ENOTDIR;

		while (sqfs_dir_next(&data->image, &dir, &entry, &err)) {
			if (err != SQFS_OK) return -EIO;
			passed_n++;
		}

		if (passed_n != 0) {
			passed = malloc(sizeof *passed * passed_n);

			sqfs_dir_open(&data->image, &inode, &dir, 0);

			while (sqfs_dir_next(&data->image, &dir, &entry, &err)) {
				if (err != SQFS_OK) {
					err = -EIO;
					goto error;
				}

				char *name = malloc(entry.name_size + 1);
				name[entry.name_size] = '\0';
				memcpy(name, entry.name, entry.name_size);
				passed[passed_i++] = name;

				struct stat st = {
					.st_ino = entry.inode_number,
					.st_mode = entry.type << 12,
				};

				if (
					#if FUSE_USE_VERSION < 30
						filler(buf, name, &st, 0)
					#else
						filler(buf, name, &st, 0, 0)
					#endif
				) goto error;
			}
		}
	}

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) {
		err = -ENAMETOOLONG;
		goto error;
	}

	DIR *root_dir = opendir(buff);
	if (!root_dir) return -errno;

	struct dirent *root_entry;

	while ((root_entry = readdir(root_dir))) {
		struct stat st = {
			.st_ino = root_entry->d_ino,
			.st_mode = root_entry->d_type << 12,
		};

		// Will be slow for overlay dirs with a lot of files, but so what?
		bool skip = false;

		for (size_t i = 0; i < passed_n; i++) {
			if (!strcmp(passed[i], root_entry->d_name)) {
				skip = true;
				break;
			}
		}

		if (skip) continue;

		if (
			#if FUSE_USE_VERSION < 30
				filler(buf, root_entry->d_name, &st, 0)
			#else
				filler(buf, entry->d_name, &st, 0, 0)
			#endif
		) {
			closedir(root_dir);
			goto error;
		}
	}

	closedir(root_dir);

error:
	if (passed) {
		for (size_t i = 0; i < passed_i; i++) free(passed[i]);
		free(passed);
	}

	return err;
}
static int slimfs_getattr(
	#if FUSE_USE_VERSION < 30
		const char *path, struct stat *stat
	#else
		const char *path, struct stat *stat, struct fuse_file_info *fi
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS GETATTR %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// The overlay FS can't overlay the root dir
	if (strcmp(path, "/") && find_squashfs_node(&data->image, path, &inode)) {
		memset(stat, 0, sizeof(*stat));
		stat->st_ino = inode.base.inode_number;
		stat->st_mode = inode.base.mode;
		stat->st_nlink = inode.nlink;
		stat->st_mtime = stat->st_ctime = stat->st_atime = inode.base.mtime;

		if (S_ISREG(stat->st_mode)) {
			// FIXME: do symlinks, dirs, etc have a size?
			stat->st_size = inode.xtra.reg.file_size;
			stat->st_blocks = stat->st_size / 512;
		}
		else if (S_ISBLK(stat->st_mode) || S_ISCHR(stat->st_mode)) {
			stat->st_rdev = makedev(inode.xtra.dev.major, inode.xtra.dev.minor);
		}
		else if (S_ISLNK(stat->st_mode)) {
			stat->st_size = inode.xtra.symlink_size;
		}

		stat->st_blksize = data->image.sb.block_size;

		// if (data->image.uid > 0) {
		stat->st_uid = data->image.uid;
		// }
		// else {
		// 	uid_t id;
		// 	fprintf(stderr, "HELL NO!");
		// 	if (sqfs_id_get(&data->image, inode.base.uid, &id) != SQFS_OK) return -EIO;
		// 	stat->st_uid = id;
		// }

		// if (data->image.gid > 0) {
		stat->st_gid = data->image.gid;
		// }
		// else {
		// 	uid_t id;
		// 	fprintf(stderr, "HELL NO!");
		// 	if (sqfs_id_get(&data->image, inode.base.guid, &id) != SQFS_OK) return -EIO;
		// 	stat->st_gid = id;
		// }

		return 0;
	}
	else if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;
	else {
		if (lstat(buff, stat) < 0) return -errno;
	}

	// This is a workaround for broken gnu find implementations. Actually,
	// n_links is not defined at all for directories by posix. However, it
	// seems to be common for filesystems to set it to one if the actual value
	// is unknown. Since nlink_t is unsigned and since these broken implementations
	// always substract 2 (for . and ..) this will cause an underflow, setting
	// it to max(nlink_t).
	if (S_ISDIR(stat->st_mode)) stat->st_nlink = 1;
	return 0;
}
static int slimfs_access(const char *path, int mask) {
	if (DEBUG) fprintf(stderr, "SLIMFS ACCESS %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	if (find_squashfs_node(&data->image, path, &inode)) {
		// The squashfs image is allowed to specify the executable perms
		if (mask & X_OK) {
			if ((inode.base.mode & S_IXUSR) | (inode.base.mode & S_IXGRP) | (inode.base.mode & S_IXOTH)) return 0;
			return -EACCES;
		}
		// The image is always read-only
		if (mask & W_OK) return -EACCES;
		// Although the image can specify no read perms, we will ignore those
		if (mask & R_OK) return 0;
	}
	else if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;
	else {
		struct stat stat;
		if (lstat(buff, &stat) < 0) return -errno;

		if ((mask & X_OK) && (stat.st_mode & S_IXUSR) == 0) return -EACCES;
		if ((mask & W_OK) && (stat.st_mode & S_IWUSR) == 0) return -EACCES;
		if ((mask & R_OK) && (stat.st_mode & S_IRUSR) == 0) return -EACCES;
	}

	return 0;
}
// Wrapper function to convert the result of statfs() to statvfs()
// libfuse uses statvfs, since it conforms to POSIX. Unfortunately,
// glibc's statvfs parses /proc/mounts, which then results in reading
// the filesystem itself again - which would result in a deadlock.
// TODO: BSD/MacOSX
static int statvfs_local(const char *path, struct statvfs *stbuf) {
#ifdef linux
	// glibc's statvfs walks /proc/mounts and stats entries found there
	// in order to extract their mount flags, which may deadlock if they
	// are mounted under the slimfs. As a result, we have to do this ourselves.
	struct statfs stfs;
	int res = statfs(path, &stfs);
	if (res == -1) return res;

	*stbuf = (struct statvfs){
		.f_bsize = stfs.f_bsize,
		.f_frsize = stfs.f_frsize ? stfs.f_frsize : stfs.f_bsize,
		.f_blocks = stfs.f_blocks,
		.f_bfree = stfs.f_bfree,
		.f_bavail = stfs.f_bavail,
		.f_files = stfs.f_files,
		.f_ffree = stfs.f_ffree,
		// nobody knows
		.f_favail = stfs.f_ffree,
		// We don't worry about flags, exactly because this would
		// require reading /proc/mounts, and avoiding that and the
		// resulting deadlocks is exactly what we're trying to avoid
		// by doing this rather than using statvfs.
		.f_flag = 0,
		.f_namemax = stfs.f_namelen,
	};


	return 0;
#else
	return statvfs(path, stbuf);
#endif
}
// statfs implementation
// Note: We do not set the fsid, as fuse ignores it anyway.
static int slimfs_statfs(const char *path, struct statvfs *stbuf) {
	if (DEBUG) fprintf(stderr, "SLIMFS STATFS %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	// We could do something more sophisticated OR we could just stat the root

	int res = statvfs_local(data->root_dir, stbuf);
	if (res < 0) return -errno;
	return 0;

	// (void)path;

	// int first = 1;

	// dev_t devno[uopt.nbranches];

	// int retVal = 0;

	// int i = 0;
	// for (i = 0; i < uopt.nbranches; i++) {
	// 	struct statvfs stb;
	// 	int res = statvfs_local(uopt.branches[i].path, &stb);
	// 	if (res == -1) {
	// 		retVal = -errno;
	// 		break;
	// 	}

	// 	struct stat st;
	// 	res = stat(uopt.branches[i].path, &st);
	// 	if (res == -1) {
	// 		retVal = -errno;
	// 		break;
	// 	}
	// 	devno[i] = st.st_dev;

	// 	if (first) {
	// 		memcpy(stbuf, &stb, sizeof(*stbuf));
	// 		first = 0;
	// 		stbuf->f_fsid = stb.f_fsid << 8;
	// 		continue;
	// 	}

	// 	// Eliminate same devices
	// 	int j = 0;
	// 	for (j = 0; j < i; j ++) {
	// 		if (st.st_dev == devno[j]) break;
	// 	}

	// 	if (j == i) {
	// 		// Filesystem can have different block sizes -> normalize to first's block size
	// 		double ratio = (double)stb.f_bsize / (double)stbuf->f_bsize;

	// 		if (uopt.branches[i].rw) {
	// 			stbuf->f_blocks += stb.f_blocks * ratio;
	// 			stbuf->f_bfree += stb.f_bfree * ratio;
	// 			stbuf->f_bavail += stb.f_bavail * ratio;

	// 			stbuf->f_files += stb.f_files;
	// 			stbuf->f_ffree += stb.f_ffree;
	// 			stbuf->f_favail += stb.f_favail;
	// 		}
	// 		else if (!uopt.statfs_omit_ro) {
	// 			// omitting the RO branches is not correct regarding
	// 			// the block counts but it actually fixes the
	// 			// percentage of free space. so, let the user decide.
	// 			stbuf->f_blocks += stb.f_blocks * ratio;
	// 			stbuf->f_files  += stb.f_files;
	// 		}

	// 		if (!(stb.f_flag & ST_RDONLY)) stbuf->f_flag &= ~ST_RDONLY;
	// 		if (!(stb.f_flag & ST_NOSUID)) stbuf->f_flag &= ~ST_NOSUID;

	// 		if (stb.f_namemax < stbuf->f_namemax) stbuf->f_namemax = stb.f_namemax;
	// 	}
	// }

	// return retVal;
}
static int slimfs_utimens(
	#if FUSE_USE_VERSION < 30
		const char *path, const struct timespec ts[2]
	#else
		const char *path, const struct timespec ts[2], struct fuse_file_info *fi
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS UTIMENS %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	if (find_squashfs_node(&data->image, path, &inode)) return -EROFS;
	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	if (utimensat(0, buff, ts, AT_SYMLINK_NOFOLLOW) < 0) return -errno;
	else return 0;
}

static int slimfs_chmod(
	#if FUSE_USE_VERSION < 30
		const char *path, mode_t mode
	#else
		const char *path, mode_t mode, struct fuse_file_info *fi
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS CHMOD %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// If the file is in the overlay, we can't touch it
	if (find_squashfs_node(&data->image, path, &inode)) return -EROFS;

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	int res = chmod(buff, mode);
	if (res == -1) return -errno;

	return 0;
}
static int slimfs_chown(
	#if FUSE_USE_VERSION < 30
		const char *path, uid_t uid, gid_t gid
	#else
		const char *path, uid_t uid, gid_t gid, struct fuse_file_info *fi
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS CHOWN %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// If the file is in the overlay, we can't touch it
	if (find_squashfs_node(&data->image, path, &inode)) return -EROFS;

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	int res = lchown(buff, uid, gid);
	if (res == -1) return -errno;

	return 0;
}

// Stub function, look into removing it
static int slimfs_ioctl(
	#if FUSE_USE_VERSION < 35
		const char *path, int cmd, void *arg, struct fuse_file_info *fi, unsigned int flags, void *data
	#else
		const char *path, unsigned int cmd, void *arg, struct fuse_file_info *fi, unsigned int flags, void *data
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS IOCTL %s\n", path);

	#ifdef FUSE_IOCTL_COMPAT
		if (flags & FUSE_IOCTL_COMPAT) return -ENOSYS;
	#endif

	// usyslog(LOG_ERR, "Unknown ioctl: %d", cmd);
	return -EINVAL;
}

static int slimfs_truncate(
	#if FUSE_USE_VERSION < 30
		const char *path, off_t size
	#else
		const char *path, off_t size, struct fuse_file_info *fi
	#endif
) {
	if (DEBUG) fprintf(stderr, "SLIMFS TRUNCATE %s\n", path);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	// Can't truncate in overlay
	if (find_squashfs_node(&data->image, path, &inode)) return -EROFS;

	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

	if (truncate(buff, size) < 0) return -errno;
	return 0;
}

static int slimfs_open(const char *path, struct fuse_file_info *fi) {
	if (DEBUG) fprintf(stderr, "SLIMFS OPEN %s (flags %d)\n", path, fi->flags);

	slimfs_t data = fuse_get_context()->private_data;
	struct sqfs_inode inode;
	char buff[PATH_MAX];

	if (find_squashfs_node(&data->image, path, &inode)) {
		// Images in the overlay may be opened only in read-only mode
		if (fi->flags & (O_WRONLY | O_RDWR)) return -EACCES;
		if (!S_ISREG(inode.base.mode)) return -EINVAL;

		user_fd_t *res_fd = malloc(sizeof *res_fd);
		res_fd->is_sqfs = true;
		res_fd->sqfs_fd = inode;

		fi->fh = (uint64_t)res_fd;
		return 0;
	}
	else {
		if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

		int res = open(buff, fi->flags);
		if (res < 0) return -errno;

		user_fd_t *res_fd = malloc(sizeof *res_fd);
		res_fd->is_sqfs = false;
		res_fd->real_fd = res;

		fi->fh = (uint64_t)res_fd;
		return 0;
	}
}
static int slimfs_release(const char *path, struct fuse_file_info *fi) {
	if (DEBUG) fprintf(stderr, "SLIMFS CLOSE %s\n", path);

	user_fd_t *fd = (user_fd_t*)fi->fh;

	if (fd->is_sqfs) {
		free(fd);
		return 0;
	}
	else {
		int err = close(fd->real_fd);
		free(fd);

		if (err < 0) return -errno;
		return 0;
	}
}

static int slimfs_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
	if (DEBUG) fprintf(stderr, "SLIMFS READ %s (%lu - %lu)", path, offset, offset + size);

	user_fd_t *fd = (user_fd_t*)fi->fh;

	if (fd->is_sqfs) {
		slimfs_t data = fuse_get_context()->private_data;
		if (sqfs_read_range(&data->image, &fd->sqfs_fd, offset, (sqfs_off_t*)&size, buf) != SQFS_OK) return -EIO;
		return (int)size;
	}
	else {
		int res = pread(fd->real_fd, buf, size, offset);
		if (res == -1) return -errno;
		return res;
	}
}
static int slimfs_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
	if (DEBUG) fprintf(stderr, "SLIMFS WRITE %s (%lu - %lu)", path, offset, offset + size);

	user_fd_t *fd = (user_fd_t*)fi->fh;

	if (fd->is_sqfs) return -EACCES;

	int res = pwrite(fd->real_fd, buf, size, offset);
	if (res == -1) return -errno;
	return res;
}
static int slimfs_flush(const char *path, struct fuse_file_info *fi) {
	if (DEBUG) fprintf(stderr, "SLIMFS FLUSH %s\n", path);

	user_fd_t *fd = (user_fd_t*)fi->fh;

	if (fd->is_sqfs) {
		// We will pretend we have flushed the written data
		// (we can't fail this operation since we aren't doing anything)
		return 0;
	}
	else {
		if (fsync(fd->real_fd) < 0) return -errno;
		else return 0;
	}

	// int fd = dup(fi->fh);

	// if (fd == -1) {
	// 	// What to do now?
	// 	if (fsync(fi->fh) == -1) return -EIO;
	// 	return -errno;
	// }

	// int res = close(fd);
	// if (res == -1) return -errno;

	// return 0;
}

// static int slimfs_getxattr(
// 	#if __APPLE__
// 		const char *path, const char *name, char *value, size_t size, uint32_t position
// 	#else
// 		const char *path, const char *name, char *value, size_t size
// 	#endif
// ) {
// 	slimfs_t data = fuse_get_context()->private_data;
// 	char buff[PATH_MAX];

// 	if (!find_file(buff, sizeof buff, data->overlay_dir, path)) {
// 		if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;
// 	}

// 	#if __APPLE__
// 		int res = getxattr(buff, name, value, size, position, XATTR_NOFOLLOW);
// 	#else
// 		int res = lgetxattr(buff, name, value, size);
// 	#endif

// 	if (res < 0) return -errno;
// 	return 0;
// }
// static int slimfs_listxattr(const char *path, char *list, size_t size) {
// 	slimfs_t data = fuse_get_context()->private_data;
// 	char buff[PATH_MAX];

// 	if (!find_file(buff, sizeof buff, data->overlay_dir, path)) {
// 		if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;
// 	}

// 	#if __APPLE__
// 		int res = listxattr(buff, list, size, XATTR_NOFOLLOW);
// 	#else
// 		int res = llistxattr(buff, list, size);
// 	#endif

// 	if (res < 0) return -errno;
// 	return 0;
// }
// static int slimfs_removexattr(const char *path, const char *name) {
// 	slimfs_t data = fuse_get_context()->private_data;
// 	char buff[PATH_MAX];

// 	if (find_file(buff, sizeof buff, data->overlay_dir, path)) return -EROFS;
// 	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

// 	#if __APPLE__
// 		int res = removexattr(buff, name, XATTR_NOFOLLOW);
// 	#else
// 		int res = lremovexattr(buff, name);
// 	#endif

// 	if (res < 0) return -errno;
// 	return 0;
// }
// static int slimfs_setxattr(
// 	#if __APPLE__
// 		const char *path, const char *name, const char *value, size_t size, int flags, uint32_t position
// 	#else
// 		const char *path, const char *name, const char *value, size_t size, int flags
// 	#endif
// ) {
// 	slimfs_t data = fuse_get_context()->private_data;
// 	char buff[PATH_MAX];

// 	if (find_file(buff, sizeof buff, data->overlay_dir, path)) return -EROFS;
// 	if (!cat_path(buff, sizeof buff, data->root_dir, path)) return -ENAMETOOLONG;

// 	#if __APPLE__
// 		int res = setxattr(buff, name, value, size, position, (flags | XATTR_NOFOLLOW) & ~XATTR_NOSECURITY);
// 	#else
// 		int res = lsetxattr(buff, name, value, size, flags);
// 	#endif

// 	if (res < 0) return -errno;
// 	return 0;
// }

struct fuse_operations slimfs_oper = {
	.chmod = slimfs_chmod,
	.chown = slimfs_chown,
	.create = slimfs_create,
	.flush = slimfs_flush,
	.getattr = slimfs_getattr,
	.access = slimfs_access,
	.init = slimfs_init,
	.destroy = slimfs_destroy,
	.ioctl = slimfs_ioctl,
	.link = slimfs_link,
	.mkdir = slimfs_mkdir,
	.mknod = slimfs_mknod,
	.open = slimfs_open,
	.read = slimfs_read,
	.readlink = slimfs_readlink,
	.readdir = slimfs_readdir,
	.release = slimfs_release,
	.rename = slimfs_rename,
	.rmdir = slimfs_rmdir,
	.statfs = slimfs_statfs,
	.symlink = slimfs_symlink,
	.truncate = slimfs_truncate,
	.unlink = slimfs_unlink,
	.utimens = slimfs_utimens,
	.write = slimfs_write,
	// .getxattr = slimfs_getxattr,
	// .listxattr = slimfs_listxattr,
	// .removexattr = slimfs_removexattr,
	// .setxattr = slimfs_setxattr,
};

slimfs_t slimfs_new(int argc, char **argv, const char *target_dir, const char *root_dir, const char *img_path, size_t img_offset) {
	struct sqfs sqfs;
	if (sqfs_open_image(&sqfs, img_path, img_offset) != SQFS_OK) {
		warnx("Failed to open squashfs image");
		return NULL;
	}

	struct fuse_args args = FUSE_ARGS_INIT(0, NULL);
	for (int i = 0; i < argc; i++) {
		if (!argv[i]) break;
		fuse_opt_add_arg(&args, argv[i]);
	}

	#ifdef FUSE_CAP_BIG_WRITES
		fuse_opt_add_arg(&args, "-obig_writes");
	#endif

	struct fuse_chan *channel = fuse_mount(target_dir, &args);
	if (!channel) {
		warnx("Failed to create fuse channel");
		fuse_opt_free_args(&args);
		sqfs_destroy(&sqfs);
		return NULL;
	}

	slimfs_t data = malloc(sizeof *data);
	if (strlen(root_dir) <= 0) root_dir = ".";
	strncpy(data->root_dir, root_dir, sizeof data->root_dir);

	size_t len = strlen(data->root_dir) - 1;
	for (size_t i = len; i > 0; i--) {
		if (data->root_dir[len - 1] != '/') break;
		data->root_dir[len - 1] = '\0';
	}

	strncpy(data->target_dir, target_dir, sizeof data->target_dir);
	data->image = sqfs;

	struct fuse *fuse = fuse_new(channel, &args, &slimfs_oper, sizeof slimfs_oper, data);
	if (!fuse) {
		warnx("Failed to create fuse");
		fuse_destroy(fuse);
		fuse_unmount(target_dir, channel);
		fuse_opt_free_args(&args);
		sqfs_destroy(&sqfs);
		return NULL;
	}

	fuse_opt_free_args(&args);

	data->fuse = fuse;
	data->channel = channel;

	return data;
}

void slimfs_free(slimfs_t fs) {
	fuse_destroy(fs->fuse);
}
