local ostree = require "ostree";
local flatpak = require "flatpak";
local fs = require "fs";
local elf_loader = require "elf_loader";

local entry_preifx = [[#!/bin/sh

export PATH=$PATH:/app/bin
]];

local function help()
	print "Slimpack pack help:";
	print "Subcommand of slimpack, used to repack a flatpak application into";
	print "an AppImage-like self-contained executable.";
	print "  --help: Shows this message";
	print "  --target (-o): Specify the output file. Defaults to 'app'";
	print "  <arg>: Specifies a package to be converted. At the moment, one may be specified.";
	print "         in the future however, extensions will be specified as more package refs";

	os.exit();
end

local function dump_at(repo, name, target)
	local ref = repo:read_ref(name);
	local commit = repo:read_commit(ref);

	assert(repo:read_dir(commit.root))
		:dir(repo, "files")
		:recurse(repo, function (path, node)
			print(path);
			return ostree.dump_node(target .. path, node);
		end);
end

local function write_entry(repo, name, target)
	local meta = flatpak.parse_package(repo, name);

	local f = assert(io.open(target, "wb"));
	f:write(entry_preifx);
	f:write(meta.command);
	f:close();
	assert(fs.chmod(target, "755"));
end

local function write_res(img_path, target)
	local f = assert(io.open(target, "wb"));
	f:write(elf_loader);

	local img_f = assert(io.open(img_path, "rb"));

	for part in img_f:lines(4096) do
		f:write(part);
	end

	img_f:close();
	f:close();
	assert(fs.chmod(target, "755"));
end

return function (options, ...)
	local target;
	local name;

	require "util.args" ({
		function (val)
			if name then
				error "only one package may be specified (at the moment)";
			else
				name = val;
				return true;
			end
		end,
		target = function (val)
			if target then
				error "target may be specified only once";
			else
				target = val;
				return true;
			end
		end,
		help = help,

		o = "target",
	}, ...);

	target = target or "./app";
	if not name then
		error "package must be specified";
	end

	local repo = ostree.from_http(options.repo);

	local tmp_dir = assert(io.popen "mktemp -d .pack_XXXXXXXX":read "*a"):sub(1, -2);

	local ok, err = pcall(function ()
		dump_at(repo, name, tmp_dir .. "/dump/app");

		write_entry(repo, name, tmp_dir .. "/dump/entry");

		-- TODO: replace with C code
		assert(os.execute(("mksquashfs %q %q"):format(tmp_dir .. "/dump", tmp_dir .. "/img.squashfs")));

		write_res(tmp_dir .. "/img.squashfs", target);
	end);

	-- TODO: replace with C code
	os.execute(("rm -rf %q"):format(tmp_dir));

	if not ok then error(err, 0) end
end
