local gvariant = require "formats.gvariant";
local fmt = require "fmt";
local zlib = require "zlib";
local fs = require "fs";

local header_fmt = fmt.new ">! I4 I4";

--- OSTREE_COMMIT_GVARIANT_FORMAT:
--- - a{sv}: metadata
--- - r: previous commit checksum
--- - a(sr): related objects
--- - s: subject
--- - s: body
--- - t: timestamp
--- - r: root dir ref
--- - r: root dirmeta ref
local read_commit = gvariant "a{sv}ra(sr)sstrr";

--- OSTREE_TREE_GVARIANT_FORMAT:
--- - a(sr): a map of filename to its file ref
--- - a(srr): a map of dirname to its dir ref and dirmeta ref
local read_dir = gvariant "a(sr)a(srr)";

--- OSTREE_DIRMETA_GVARIANT_FORMAT:
--- - u: uid (big-endian)
--- - u: gid (big-endian)
--- - u: mode (big-endian)
--- - a(ayay): xattrs
local read_dirmeta = gvariant "uuua(ayay)";

--- - t: size
--- - u: uid
--- - u: gid
--- - u: mode
--- - u: rdev
--- - s: symlink_target
--- - a(ayay): xattrs
local read_header = gvariant "tuuuusa(ayay)";

--- OSTREE_SUMMARY_GVARIANT_FORMAT:
--- - a{s(tra{sv})}: map of ref name to data about it
--- 	- t: commit size
--- 	- r: checksum
--- 	- a{sv}: metadata
--- - a{sv}: metadata
local read_summary = gvariant "l(s(tra{sv}))a{sv}";

--- Flatpak-specific format (I think):
--- - a{s(rara{sv})}: map of architecture name to data about it
--- 	- r: checksum
--- 	- ar: history of previous checksums
--- 	- a{sv}: metadata
--- - a{sv}: metadata
local read_summary_index = gvariant "a{s(rara{sv})}a{sv}";

--- @class ostree_file
--- @field type "file"
--- @field uid integer
--- @field gid integer
--- @field mode integer
--- @field size integer
--- @field get_content fun(): fun(): string?

--- @class ostree_link
--- @field type "link"
--- @field uid integer
--- @field gid integer
--- @field mode integer
--- @field target string

--- @class ostree_dir
--- @field type "dir"
--- @field files { [string]: string }
--- @field dirs { [string]: { [1]: string, [2]: string } }
--- @field uid integer
--- @field gid integer
--- @field mode integer
local ostree_dir = {};
ostree_dir.__index = ostree_dir;

--- @param repo ostree
--- @param name string
function ostree_dir:file(repo, name)
	if self.files[name] == nil then return nil end
	return repo:read_file(self.files[name]);
end
--- @param repo ostree
--- @param name string
function ostree_dir:dir(repo, name)
	if self.dirs[name] == nil then return nil end
	return repo:read_dir(self.dirs[name]);
end
--- @param repo ostree
--- @param callback fun(path: string, val: ostree_dir | ostree_file | ostree_link)
function ostree_dir:recurse(repo, callback)
	local function recurser(self, repo, callback, path)
		callback(path, self);

		for file, ref in pairs(self.files) do
			callback(path .. file, repo:read_file(ref));
		end

		for dir, ref in pairs(self.dirs) do
			recurser(repo:read_dir(ref), repo, callback, path .. dir .. "/");
		end
	end

	return recurser(self, repo, callback, "/");
end

--- @class ostree
--- @field reader fun(path: string): string
local ostree = {};
ostree.__index = ostree;

--- @param reader fun(path: string): string
--- @return ostree
function ostree.new(reader)
	return setmetatable({ reader = reader }, ostree);
end
--- @param url string
--- @return ostree
function ostree.from_http(url)
	local http = require "http";

	return ostree.new(function (path)
		return http.get(url .. "/" .. path);
	end);
end
--- @param root string
--- @return ostree
function ostree.from_fs(root)
	return ostree.new(function (path)
		local f = assert(io.open(root .. "/" .. path));
		local content = assert(f:read "*a");
		assert(f:close());

		return content;
	end);
end

--- @param data string
function ostree.parse_commit(data)
	local metadata, prev, related, subject, body, timestamp, root_dir, rood_dirmeta = read_commit(data);

	return {
		metadata = metadata,
		prev = prev,
		related = related,
		subject = subject,
		body = body,
		timestamp = timestamp,
		root = { root_dir, rood_dirmeta },
	};
end
--- @param dir string | { [1]: string, [2]?: string }
--- @param meta? string
function ostree.parse_dir(dir, meta)
	if type(dir) ~= "string" then
		meta = dir[2];
		dir = dir[1];
	end

	local files_arr, dirs_arr = read_dir(dir);

	local files = {};
	local dirs = {};

	for i = 1, #files_arr do
		files[files_arr[i][1]] = files_arr[i][2];
	end
	for i = 1, #dirs_arr do
		dirs[dirs_arr[i][1]] = { dirs_arr[i][2], dirs_arr[i][3] };
	end

	if meta then
		local uid, gid, mode, xattrs = read_dirmeta(meta);
		return setmetatable({ type = "dir", files = files, dirs = dirs, uid = uid, gid = gid, mode = mode & 0xFFF, xattrs = xattrs }, ostree_dir);
	else
		return setmetatable({ type = "dir", files = files, dirs = dirs }, ostree_dir);
	end
end
--- @param data string
--- @return ostree_file
function ostree.parse_file(data)
	local header_size = header_fmt:unpack(data);
	local header = data:sub(9, header_size + 8);
	local size, uid, gid, mode, rdev, link, xattrs = read_header(header);

	if link ~= "" then
		if size > 0 then
			error "link may not have contents";
		else
			return {
				type = "link",
				target = link,
				uid = uid,
				gid = gid,
				mode = mode & 0xFFF,
				xattrs = xattrs,
			};
		end
	else
		local get_content;
		if size == 0 then
			function get_content()
				return function ()
					return nil;
				end
			end
		else
			function get_content()
				local compressed = data:sub(header_size + 9);
				return zlib.inflate_raw(compressed);
			end
		end

		return {
			type = "file",
			get_content = get_content,
			uid = uid,
			gid = gid,
			mode = mode & 0xFFF,
			size = size,
			xattrs = xattrs,
		};
	end
end
--- @param data string
function ostree.parse_summary(data)
	local refs, metadata = read_summary(data);

	local function refs_iter()
		local el = refs();

		if el == nil then
			return nil;
		else
			return el[1], {
				size = el[2][1],
				checksum = el[2][2],
				metadata = el[2][3],
			};
		end
	end

	return {
		refs = refs_iter,
		metadata = metadata,
	};
end
--- @param data string
function ostree.parse_subsummary(data)
	local parts = {};
	for part in zlib.gunzip(data) do parts[#parts + 1] = part end
	return ostree.parse_summary(table.concat(parts));
end
--- @param data string
function ostree.parse_summary_index(data)
	local refs, metadata = read_summary_index(data);

	for key, val in pairs(refs) do
		refs[key] = {
			checksum = val[1],
			-- history = val[2],
			metadata = val[3],
		};
	end

	return {
		refs = refs,
		metadata = metadata,
	};
end
--- @param path string
--- @param node ostree_dir | ostree_link | ostree_file
function ostree.dump_node(path, node)
	if node.type == "dir" then
		assert(fs.mkdir(path, true));
		assert(fs.chmod(path, node.mode & 0xFFF));
	elseif node.type == "file" then
		assert(fs.mkdir(path:match "^(.+)/[^/]+$", true));

		local f = assert(io.open(path, "w"));

		for seg in node.get_content() do
			assert(f:write(seg));
		end

		assert(f:close());

		assert(fs.chmod(path, node.mode & 0xFFF));
	elseif node.type == "link" then
		assert(fs.mkdir(path:match "^(.+)/[^/]+$", true));
		assert(fs.symlink(node.target, path));
	else
		error("Unknown node type '" .. node.type .. "'");
	end
end

--- @param type string
--- @param ref string
function ostree:read_object(type, ref)
	return self.reader("objects/" .. ref:sub(1, 2) .. "/" .. ref:sub(3) .. "." .. type);
end

--- @param ref string
function ostree:read_commit(ref)
	return ostree.parse_commit(self:read_object("commit", ref));
end
--- @param dir string | { [1]: string, [2]?: string }
--- @param meta? string
function ostree:read_dir(dir, meta)
	if type(dir) ~= "string" then
		meta = dir[2];
		dir = dir[1];
	end

	return ostree.parse_dir(self:read_object("dirtree", dir), meta and self:read_object("dirmeta", meta));
end
--- @param ref string
function ostree:read_file(ref)
	return ostree.parse_file(self:read_object("filez", ref));
end
function ostree:read_summary()
	print(self.reader("summary"));
	return ostree.parse_summary(self.reader("summary"));
end
function ostree:read_subsummary(ref)
	return ostree.parse_subsummary(self.reader("summaries/" .. ref .. ".gz"));
end
function ostree:read_summary_index()
	return ostree.parse_summary_index(self.reader("summary.idx"));
end
--- @param pkg string
function ostree:read_ref(pkg)
	return self.reader("refs/heads/" .. pkg):gsub("%s", "");
end

return ostree;
